const I = require('immutable-ext')
const Task = require('data.task')
const Either = require('data.either')
const Kleisli = require('../../lib/kleisli')
const {variants} = require('../../lib/finds')
const {setDefaultMods, applyModsToSelectionOrAll} = require('../helpers/modifiers')
const {getMarkup} = require('../helpers/markup')
const { createValidator } = require('@salesforce-ux/design-system-markup/server')

module.exports = (parser, delegates) => {
  const validate = createValidator(parser.comments.map(cs => cs.get('annotations')))

  // setNewComponent :: Map({component :: String}) -> State -> Task Err State
  const setNewComponent = payload => state =>
    parser
      .component(payload.get('component') || state.get('component').get('id'))
      .fold(() => Task.rejected('could not find component'), Task.of)
      .map(component => state
        .set('component', component)
        .set('variant', variants(component).first())
        .set('modifiers', I.Set())
        .set('currentSelection', null)
      )
      .chain(state =>
        state.get('variant')
        ? setExample(state)
        : Task.of(state)
      )

  const order = ['default', 'states', 'examples']
  const markupOrder = x => order.indexOf(x.get('title'))

  // setExample :: State -> Task Err State
  const setExample = state =>
    getMarkup(delegates.getMarkup, state)
    .map(sections => sections.sortBy(markupOrder))
    .map(sections => applyModsToSelectionOrAll(state, sections))
    .map(({modifiers, sections}) =>
      state
      .set('modifiers', modifiers)
      .update('markup', () =>
        sections.map(section =>
          section.update('items', items =>
            items.map(item =>
              item.set('errors', validate(item.get('markup')))
            )
          )
        )
      )
    )

  // setModifiersFromString :: Map({modifiers :: String}) -> State -> Task Err State
  const setModifiersFromString = payload => state =>
    Either.fromNullable(payload.get('modifiers'))
    .map(x => I.List.of(x.split(',')))
    .chain(names =>
      names.traverse(Either.of, x => parser.findModifier(state.get('component'), x)))
    .fold(() => Task.of(I.List()), Task.of)
    .map(mods => state.set('modifiers', mods))

  // setVariantFromId :: Map({variant :: String}) -> State -> Task Err State
  const setVariantFromId = payload => state =>
    Either
    .of(state)
    .chain(state =>
      parser
      .findVariant(state.get('component'), payload.get('variant'))
      .map(variant =>
        state
        .set('variant', variant)
      )
    )
    .fold(() => Task.rejected('could not set variant'), Task.of)
    .map(state =>
      state.set('modifiers', setDefaultMods(state))
    )
    .chain(setExample)

  const setAllValues = payload => state =>
    I.List
    .of(setModifiersFromString, setVariantFromId, setNewComponent)
    .foldMap(f => Kleisli(f(payload)), Kleisli(Task.of))
    .fold(state)

  // reducer :: (State, {type: String, payload: a}) -> Task(State)
  const reducer = (state, action) => {
    const type = action.get('type')
    const payload = action.get('payload')
    switch (type) {
      case 'initialize':
        return Task
               .of(parser.components())
               .map(cs =>
                  I.Map({
                    components: cs.map(c => c.get('id')),
                    formFactor: 'desktop',
                    bgColor: 'white',
                    utilities: parser
                               .utilities()
                               .traverse(Either.of, u =>
                                 parser.utility(u.get('id'))).get()
                  }))
              .chain(state =>
                Either
                .fromNullable(state.get('components').first())
                .chain(parser.component)
                .map(comp => state.set('component', comp))
                .chain(st =>
                  Either
                  .fromNullable(variants(st.get('component')).first())
                  .map(v => v.get('id'))
                  .map(variant => setVariantFromId(I.Map({variant}))(st))
                )
                .getOrElse(Task.of(state))
              )
      case 'loadFromUrl':
        return setAllValues(payload)(state)
      case 'componentSelected':
        return setNewComponent(payload)(state)
      case 'variantSelected':
        return setVariantFromId(payload)(state)
      case 'utilitySelected':
        return Task
               .of(state)
               .chain(s =>
                 parser.utility(payload.get('utility'))
                 .fold(() => Task.rejected('could not set utility'),
                       u => Task.of(state.set('component', u)))
               )
               .map(st =>
                 st.set('variant', st.getIn(['component', 'restrictees']).first()))
               .chain(setExample)
      case 'modifierSelected':
        return Task.of(state)
          .map(state =>
            state.set('modifiers', payload.get('modifiers'))
          )
          .chain(setExample)
      case 'exampleSelected':
        return Task.of(state)
          .map(state =>
            state.set('currentSelection', payload.get('value'))
          )
          .map(state =>
            state.set('modifiers', setDefaultMods(state))
          )
      case 'stateChange':
        return Task.of(state)
          .map(state =>
            state.set(payload.get('key'), payload.get('value'))
          )
      case 'cssChanged':
        return Task.of(state)
          .map(s =>
            s.set('styles', Date.now())
          )
      case 'commentsChanged':
        const pay = I.Map({
          component: state.getIn(['component', 'id']),
          variant: state.getIn(['variant', 'id']),
          modifiers: state.get('modifiers').map(m => m.get('id')).join(',')
        })
        console.log(pay.toJS())
        return setAllValues(pay)(state)
      default:
        return Task.of(state)
    }
  }
  return reducer
}
